# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

YELLOW = "#facc37"
CYTO_COLORS = {
    "line": YELLOW,
    "link": "green",
    "transformer": "#4290f5",
    "node": YELLOW,
    "selected": "#e28743",
    "selected_transformer": "#0349a3",
    "substation_node": "purple",
    "open_branch": "#c9c9c9",
    "highlighted": "#a10000",
}
BACKGROUND_COLOR = "#555555"
